<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:math="http://www.w3.org/2005/xpath-functions/math"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="#all"
    xmlns:hcmc="http://hcmc.uvic.ca/ns"
    xpath-default-namespace="http://www.tei-c.org/ns/1.0"
    xmlns:tei="http://www.tei-c.org/ns/1.0"
    xmlns="http://www.w3.org/1999/xhtml"
    xmlns:xh="http://www.w3.org/1999/xhtml"
    xmlns:map="http://www.w3.org/2005/xpath-functions/map"
    version="3.0">
    <xd:doc scope="stylesheet">
        <xd:desc>
            <xd:p><xd:b>Created on:</xd:b> Mar 8, 2023</xd:p>
            <xd:p><xd:b>Author:</xd:b> mholmes</xd:p>
            <xd:p></xd:p>
        </xd:desc>
    </xd:doc>
    <xd:doc>
        <xd:desc><xd:ref name="hcmc:getLocationAlphabeticBreakPoints"/>: This is passed a single HTML 
            div which contains a placename (h3) and a sequence of divs for families, each of which contains
            ul elements of which each li is a person. The idea is to determine at what alphabetic letters
            the set should be broken up to ensure that none of its constituent pages contains more than
            a supplied threshold of names, without splitting an initial letter across two pages. The 
            return value is a sequence of upper-case alphabetic letters beginning with A, which constitute
            the beginning letters for each of the output pages that will be created. This function achieves
            its job by calling the recursive function hcmc:getNextBreakpoint.</xd:desc>
        <xd:param name="placeDiv" as="element(xh:div)">The div for the place in question.</xd:param>
        <xd:param name="threshold" as="xs:integer">The breakpoint threshold which constitutes the 
            maximum number of names that should end up on a single page.</xd:param>
        <xd:return as="xs:string+">A sequence of upper-case letters constituting the first letter
            of each constituent page that should be created, beginning with e.g. 'A'.</xd:return>
    </xd:doc>
    <xsl:function name="hcmc:getLocationAlphabeticBreakPoints" as="xs:string+">
        <xsl:param name="placeDiv" as="element(xh:div)"/>
        <xsl:param name="threshold" as="xs:integer"/>
        <xsl:variable name="candidateSequence" as="xs:string+" select="(for $i in string-to-codepoints('A')[1] to string-to-codepoints('Z')[1] return codepoints-to-string($i))"/> 
        <!--<xsl:message expand-text="yes">candidateSequence ({count($candidateSequence)}): {string-join($candidateSequence, ', ')}</xsl:message>-->
        <!--<xsl:message>This message prevents an error from Saxon?!</xsl:message>-->
        <xsl:sequence 
            select="hcmc:getNextBreakpoint($placeDiv, $threshold, 0, $candidateSequence, 0, ())"/>
    </xsl:function>
    
    <xd:doc>
        <xd:desc><xd:ref name="hcmc:getNextBreakpoint"/>: This recursive function is designed to 
            assemble a sequence of breakpoints in an alphabetic sequence that breaks into sections
            such that each sequence of (typically) alphabetic letters includes no more than the 
            breakpoint number of items.</xd:desc>
        <xd:param name="placeDiv" as="element(xh:div)">The HTML div containing the people in groups
            which are alphabetized by surname.</xd:param>
        <xd:param name="threshold" as="xs:integer">The breakpoint threshold which constitutes the 
            maximum number of names that should end up on a single page.</xd:param>
        <xd:param name="prevPos" as="xs:integer">An integer indicating the previous position 
            reached in the sequence, from which this iteration should proceed.</xd:param>
        <xd:param name="candidateSequence" as="xs:string+">The sequence of letters we must iterate
            through. This is a parameter in order that we may use this function for other types
            of sequence or in other languages.
        </xd:param>
        <xd:param name="currTotal" as="xs:integer">An integer value for the current total of
            (in this case) names which have been amassed so far in the current subsequence.</xd:param>
        <xd:param name="sequenceSoFar" as="xs:string*">The sequence of letters to which we
            add new values to assemble the final return value.</xd:param>
        <xd:return as="xs:string+">A sequence of upper-case letters constituting the first letter
            of each constituent page that should be created, beginning with 'A'.</xd:return>
    </xd:doc>
    <xsl:function name="hcmc:getNextBreakpoint" as="xs:string+">
        <xsl:param name="placeDiv" as="element(xh:div)"/>
        <xsl:param name="threshold" as="xs:integer"/>
        <xsl:param name="prevPos" as="xs:integer"/>
        <xsl:param name="candidateSequence" as="xs:string+"/>
        <xsl:param name="currTotal" as="xs:integer"/>
        <xsl:param name="sequenceSoFar" as="xs:string*"/>
        <xsl:choose>
            <!-- If we're actually on the very last letter, we need to decide
        whether to include it or not, then return the final sequence. -->
            <xsl:when test="$prevPos ge (count($candidateSequence) - 1)">
                <!--<xsl:message>Final pass through getNextBreakpoint...</xsl:message>-->
                <!--<xsl:message expand-text="yes">$prevPos = {$prevPos}; count($candidateSequence) = {count($candidateSequence)}; $sequenceSoFar = {$sequenceSoFar}</xsl:message>-->
                <xsl:choose>
                    <xsl:when test="($currTotal + count($placeDiv//xh:li[xh:span[@class='name'][starts-with(., $candidateSequence[last()])]])) gt $threshold">
                        <xsl:sequence select="($sequenceSoFar, $candidateSequence[last()])"/>
                    </xsl:when>
                    <xsl:otherwise>            
                        <xsl:sequence select="$sequenceSoFar"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:when>
            
            <!-- If we're just getting started, then we know we need the first item, and we can
           just recurse from there. -->
            <xsl:when test="empty($sequenceSoFar) and $prevPos lt 1">
                <!--<xsl:message>First pass through getNextBreakpoint...</xsl:message>-->
                <xsl:sequence select="hcmc:getNextBreakpoint($placeDiv, $threshold, 1, $candidateSequence, count($placeDiv//xh:li[xh:span[@class='name'][starts-with(., $candidateSequence[1])]]), $candidateSequence[1])"/>
            </xsl:when>
            
            <!-- Otherwise, we're in the middle, and we need to decide whether we need 
           to break and start a new batch with the next letter. -->
            <xsl:otherwise>
                <!--<xsl:message expand-text="yes">Intermediate pass through getNextBreakpoint ($prevPos = {$prevPos}; count($candidateSequence) = {count($candidateSequence)}; $sequenceSoFar = {$sequenceSoFar})</xsl:message>-->
                <xsl:variable name="nextLetter" as="xs:string" select="$candidateSequence[$prevPos + 1]"/>
                <xsl:variable name="totalForNextLetter" as="xs:integer" select="count($placeDiv//xh:li[xh:span[@class='name'][starts-with(., $nextLetter)]])"/>
                <xsl:variable name="newTotal" as="xs:integer" select="$currTotal + $totalForNextLetter"/>
                <xsl:choose>
                    <!-- If we've crossed the threshold...-->
                    <xsl:when test="($newTotal gt $threshold)">
                        <!-- This letter becomes a breakpoint. -->
                        <xsl:sequence select="hcmc:getNextBreakpoint($placeDiv, $threshold, $prevPos + 1, $candidateSequence, $totalForNextLetter, (($sequenceSoFar, if ($totalForNextLetter gt 0) then $candidateSequence[($prevPos + 1)] else ())))"/>
                    </xsl:when>
                    <xsl:otherwise>
                        <!-- Otherwise, we just move up one. -->
                        <xsl:sequence select="hcmc:getNextBreakpoint($placeDiv, $threshold, ($prevPos + 1), $candidateSequence, $newTotal, $sequenceSoFar)"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:function>
</xsl:stylesheet>